from __future__ import unicode_literals


class EditingMode(object):
    # The set of key bindings that is active.
    VI = 'VI'
    EMACS = 'EMACS'


#: Name of the search buffer.
SEARCH_BUFFER = 'SEARCH_BUFFER'

#: Name of the default buffer.
DEFAULT_BUFFER = 'DEFAULT_BUFFER'

#: Name of the system buffer.
SYSTEM_BUFFER = 'SYSTEM_BUFFER'



"""
Dummy layout. Used when somebody creates an `Application` without specifying a
`Layout`.
"""
from __future__ import unicode_literals

from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.key_binding import KeyBindings

from .containers import Window
from .controls import FormattedTextControl
from .dimension import D
from .layout import Layout

__all__ = [
    'create_dummy_layout',
]


def create_dummy_layout():
    """
    Create a dummy layout for use in an 'Application' that doesn't have a
    layout specified. When ENTER is pressed, the application quits.
    """
    kb = KeyBindings()

    @kb.add('enter')
    def enter(event):
        event.app.exit()

    control = FormattedTextControl(
        HTML('No layout specified. Press <reverse>ENTER</reverse> to quit.'),
        key_bindings=kb)
    window = Window(content=control, height=D(min=1))
    return Layout(container=window, focused_element=window)



from __future__ import unicode_literals
from .application import Application
from prompt_toolkit.input import DummyInput
from prompt_toolkit.output import DummyOutput

__all__ = [
    'DummyApplication',
]


class DummyApplication(Application):
    """
    When no :class:`.Application` is running,
    :func:`.get_app` will run an instance of this :class:`.DummyApplication` instead.
    """
    def __init__(self):
        super(DummyApplication, self).__init__(output=DummyOutput(), input=DummyInput())

    def run(self):
        raise NotImplementedError('A DummyApplication is not supposed to run.')

    def run_async(self):
        raise NotImplementedError('A DummyApplication is not supposed to run.')

    def run_system_command(self):
        raise NotImplementedError

    def suspend_to_background(self):
        raise NotImplementedError



from __future__ import unicode_literals
from prompt_toolkit.keys import Keys
from ..key_bindings import KeyBindings

__all__ = [
    'load_cpr_bindings',
]


def load_cpr_bindings():
    key_bindings = KeyBindings()

    @key_bindings.add(Keys.CPRResponse, save_before=lambda e: False)
    def _(event):
        """
        Handle incoming Cursor-Position-Request response.
        """
        # The incoming data looks like u'\x1b[35;1R'
        # Parse row/col information.
        row, col = map(int, event.data[2:-1].split(';'))

        # Report absolute cursor position to the renderer.
        event.app.renderer.report_absolute_cursor_row(row)

    return key_bindings



from __future__ import unicode_literals

from prompt_toolkit.utils import is_windows
from prompt_toolkit.renderer import Output
from prompt_toolkit.win32_types import STD_OUTPUT_HANDLE
from .vt100 import Vt100_Output
from .win32 import Win32Output

from ctypes import windll, byref
from ctypes.wintypes import DWORD

__all__ = [
    'Windows10_Output',
]

# See: https://msdn.microsoft.com/pl-pl/library/windows/desktop/ms686033(v=vs.85).aspx
ENABLE_PROCESSED_INPUT = 0x0001
ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004


class Windows10_Output(object):
    """
    Windows 10 output abstraction. This enables and uses vt100 escape sequences.
    """
    def __init__(self, stdout):
        self.win32_output = Win32Output(stdout)
        self.vt100_output = Vt100_Output(stdout, lambda: None)
        self._hconsole = windll.kernel32.GetStdHandle(STD_OUTPUT_HANDLE)

    def flush(self):
        """
        Write to output stream and flush.
        """
        original_mode = DWORD(0)

        # Remember the previous console mode.
        windll.kernel32.GetConsoleMode(self._hconsole, byref(original_mode))

        # Enable processing of vt100 sequences.
        windll.kernel32.SetConsoleMode(self._hconsole, DWORD(
            ENABLE_PROCESSED_INPUT | ENABLE_VIRTUAL_TERMINAL_PROCESSING))

        try:
            self.vt100_output.flush()
        finally:
            # Restore console mode.
            windll.kernel32.SetConsoleMode(self._hconsole, original_mode)

    def __getattr__(self, name):
        if name in ('get_size', 'get_rows_below_cursor_position',
                    'enable_mouse_support', 'disable_mouse_support',
                    'scroll_buffer_to_prompt', 'get_win32_screen_buffer_info',
                    'enable_bracketed_paste', 'disable_bracketed_paste'):
            return getattr(self.win32_output, name)
        else:
            return getattr(self.vt100_output, name)


Output.register(Windows10_Output)


def is_win_vt100_enabled():
    """
    Returns True when we're running Windows and VT100 escape sequences are
    supported.
    """
    if not is_windows():
        return False

    hconsole = windll.kernel32.GetStdHandle(STD_OUTPUT_HANDLE)

    # Get original console mode.
    original_mode = DWORD(0)
    windll.kernel32.GetConsoleMode(hconsole, byref(original_mode))

    try:
        # Try to enable VT100 sequences.
        result = windll.kernel32.SetConsoleMode(hconsole, DWORD(
            ENABLE_PROCESSED_INPUT | ENABLE_VIRTUAL_TERMINAL_PROCESSING))

        return result == 1
    finally:
        windll.kernel32.SetConsoleMode(hconsole, original_mode)



"""
`Fish-style <http://fishshell.com/>`_  like auto-suggestion.

While a user types input in a certain buffer, suggestions are generated
(asynchronously.) Usually, they are displayed after the input. When the cursor
presses the right arrow and the cursor is at the end of the input, the
suggestion will be inserted.

If you want the auto suggestions to be asynchronous (in a background thread),
because they take too much time, and could potentially block the event loop,
then wrap the :class:`.AutoSuggest` instance into a
:class:`.ThreadedAutoSuggest`.
"""
from __future__ import unicode_literals
from abc import ABCMeta, abstractmethod
from six import with_metaclass

from .eventloop import Future, run_in_executor
from .filters import to_filter

__all__ = [
    'Suggestion',
    'AutoSuggest',
    'ThreadedAutoSuggest',
    'DummyAutoSuggest',
    'AutoSuggestFromHistory',
    'ConditionalAutoSuggest',
    'DynamicAutoSuggest',
]


class Suggestion(object):
    """
    Suggestion returned by an auto-suggest algorithm.

    :param text: The suggestion text.
    """
    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return 'Suggestion(%s)' % self.text


class AutoSuggest(with_metaclass(ABCMeta, object)):
    """
    Base class for auto suggestion implementations.
    """
    @abstractmethod
    def get_suggestion(self, buffer, document):
        """
        Return `None` or a :class:`.Suggestion` instance.

        We receive both :class:`~prompt_toolkit.buffer.Buffer` and
        :class:`~prompt_toolkit.document.Document`. The reason is that auto
        suggestions are retrieved asynchronously. (Like completions.) The
        buffer text could be changed in the meantime, but ``document`` contains
        the buffer document like it was at the start of the auto suggestion
        call. So, from here, don't access ``buffer.text``, but use
        ``document.text`` instead.

        :param buffer: The :class:`~prompt_toolkit.buffer.Buffer` instance.
        :param document: The :class:`~prompt_toolkit.document.Document` instance.
        """

    def get_suggestion_future(self, buff, document):
        """
        Return a :class:`.Future` which is set when the suggestions are ready.
        This function can be overloaded in order to provide an asynchronous
        implementation.
        """
        return Future.succeed(self.get_suggestion(buff, document))


class ThreadedAutoSuggest(AutoSuggest):
    """
    Wrapper that runs auto suggestions in a thread.
    (Use this to prevent the user interface from becoming unresponsive if the
    generation of suggestions takes too much time.)
    """
    def __init__(self, auto_suggest):
        assert isinstance(auto_suggest, AutoSuggest)
        self.auto_suggest = auto_suggest

    def get_suggestion(self, buff, document):
        return self.auto_suggest.get_suggestion(buff, document)

    def get_suggestion_future(self, buff, document):
        """
        Run the `get_suggestion` function in a thread.
        """
        def run_get_suggestion_thread():
            return self.get_suggestion(buff, document)
        f = run_in_executor(run_get_suggestion_thread)
        return f


class DummyAutoSuggest(AutoSuggest):
    """
    AutoSuggest class that doesn't return any suggestion.
    """
    def get_suggestion(self, buffer, document):
        return  # No suggestion


class AutoSuggestFromHistory(AutoSuggest):
    """
    Give suggestions based on the lines in the history.
    """
    def get_suggestion(self, buffer, document):
        history = buffer.history

        # Consider only the last line for the suggestion.
        text = document.text.rsplit('\n', 1)[-1]

        # Only create a suggestion when this is not an empty line.
        if text.strip():
            # Find first matching line in history.
            for string in reversed(list(history.get_strings())):
                for line in reversed(string.splitlines()):
                    if line.startswith(text):
                        return Suggestion(line[len(text):])


class ConditionalAutoSuggest(AutoSuggest):
    """
    Auto suggest that can be turned on and of according to a certain condition.
    """
    def __init__(self, auto_suggest, filter):
        assert isinstance(auto_suggest, AutoSuggest)

        self.auto_suggest = auto_suggest
        self.filter = to_filter(filter)

    def get_suggestion(self, buffer, document):
        if self.filter():
            return self.auto_suggest.get_suggestion(buffer, document)


class DynamicAutoSuggest(AutoSuggest):
    """
    Validator class that can dynamically returns any Validator.

    :param get_validator: Callable that returns a :class:`.Validator` instance.
    """
    def __init__(self, get_auto_suggest):
        assert callable(get_auto_suggest)
        self.get_auto_suggest = get_auto_suggest

    def get_suggestion(self, buff, document):
        auto_suggest = self.get_auto_suggest() or DummyAutoSuggest()
        assert isinstance(auto_suggest, AutoSuggest)
        return auto_suggest.get_suggestion(buff, document)

    def get_suggestion_future(self, buff, document):
        auto_suggest = self.get_auto_suggest() or DummyAutoSuggest()
        assert isinstance(auto_suggest, AutoSuggest)
        return auto_suggest.get_suggestion_future(buff, document)



"""
Mappings from VT100 (ANSI) escape sequences to the corresponding prompt_toolkit
keys.
"""
from __future__ import unicode_literals
from ..keys import Keys

__all__ = [
    'ANSI_SEQUENCES',
    'REVERSE_ANSI_SEQUENCES',
]


# Mapping of vt100 escape codes to Keys.
ANSI_SEQUENCES = {
    '\x00': Keys.ControlAt,  # Control-At (Also for Ctrl-Space)
    '\x01': Keys.ControlA,  # Control-A (home)
    '\x02': Keys.ControlB,  # Control-B (emacs cursor left)
    '\x03': Keys.ControlC,  # Control-C (interrupt)
    '\x04': Keys.ControlD,  # Control-D (exit)
    '\x05': Keys.ControlE,  # Control-E (end)
    '\x06': Keys.ControlF,  # Control-F (cursor forward)
    '\x07': Keys.ControlG,  # Control-G
    '\x08': Keys.ControlH,  # Control-H (8) (Identical to '\b')
    '\x09': Keys.ControlI,  # Control-I (9) (Identical to '\t')
    '\x0a': Keys.ControlJ,  # Control-J (10) (Identical to '\n')
    '\x0b': Keys.ControlK,  # Control-K (delete until end of line; vertical tab)
    '\x0c': Keys.ControlL,  # Control-L (clear; form feed)
    '\x0d': Keys.ControlM,  # Control-M (13) (Identical to '\r')
    '\x0e': Keys.ControlN,  # Control-N (14) (history forward)
    '\x0f': Keys.ControlO,  # Control-O (15)
    '\x10': Keys.ControlP,  # Control-P (16) (history back)
    '\x11': Keys.ControlQ,  # Control-Q
    '\x12': Keys.ControlR,  # Control-R (18) (reverse search)
    '\x13': Keys.ControlS,  # Control-S (19) (forward search)
    '\x14': Keys.ControlT,  # Control-T
    '\x15': Keys.ControlU,  # Control-U
    '\x16': Keys.ControlV,  # Control-V
    '\x17': Keys.ControlW,  # Control-W
    '\x18': Keys.ControlX,  # Control-X
    '\x19': Keys.ControlY,  # Control-Y (25)
    '\x1a': Keys.ControlZ,  # Control-Z

    '\x1b': Keys.Escape,            # Also Control-[
    '\x1c': Keys.ControlBackslash,  # Both Control-\ (also Ctrl-| )
    '\x1d': Keys.ControlSquareClose,  # Control-]
    '\x1e': Keys.ControlCircumflex,  # Control-^
    '\x1f': Keys.ControlUnderscore,  # Control-underscore (Also for Ctrl-hyphen.)

    # ASCII Delete (0x7f)
    # Vt220 (and Linux terminal) send this when pressing backspace. We map this
    # to ControlH, because that will make it easier to create key bindings that
    # work everywhere, with the trade-off that it's no longer possible to
    # handle backspace and control-h individually for the few terminals that
    # support it. (Most terminals send ControlH when backspace is pressed.)
    # See: http://www.ibb.net/~anne/keyboard.html
    '\x7f': Keys.ControlH,

    '\x1b[A': Keys.Up,
    '\x1b[B': Keys.Down,
    '\x1b[C': Keys.Right,
    '\x1b[D': Keys.Left,
    '\x1b[H': Keys.Home,
    '\x1bOH': Keys.Home,
    '\x1b[F': Keys.End,
    '\x1bOF': Keys.End,
    '\x1b[3~': Keys.Delete,
    '\x1b[3;2~': Keys.ShiftDelete,  # xterm, gnome-terminal.
    '\x1b[3;5~': Keys.ControlDelete,  # xterm, gnome-terminal.
    '\x1b[1~': Keys.Home,  # tmux
    '\x1b[4~': Keys.End,  # tmux
    '\x1b[5~': Keys.PageUp,
    '\x1b[6~': Keys.PageDown,
    '\x1b[7~': Keys.Home,  # xrvt
    '\x1b[8~': Keys.End,  # xrvt
    '\x1b[Z': Keys.BackTab,  # shift + tab
    '\x1b[2~': Keys.Insert,

    '\x1bOP': Keys.F1,
    '\x1bOQ': Keys.F2,
    '\x1bOR': Keys.F3,
    '\x1bOS': Keys.F4,
    '\x1b[[A': Keys.F1,  # Linux console.
    '\x1b[[B': Keys.F2,  # Linux console.
    '\x1b[[C': Keys.F3,  # Linux console.
    '\x1b[[D': Keys.F4,  # Linux console.
    '\x1b[[E': Keys.F5,  # Linux console.
    '\x1b[11~': Keys.F1,  # rxvt-unicode
    '\x1b[12~': Keys.F2,  # rxvt-unicode
    '\x1b[13~': Keys.F3,  # rxvt-unicode
    '\x1b[14~': Keys.F4,  # rxvt-unicode
    '\x1b[15~': Keys.F5,
    '\x1b[17~': Keys.F6,
    '\x1b[18~': Keys.F7,
    '\x1b[19~': Keys.F8,
    '\x1b[20~': Keys.F9,
    '\x1b[21~': Keys.F10,
    '\x1b[23~': Keys.F11,
    '\x1b[24~': Keys.F12,
    '\x1b[25~': Keys.F13,
    '\x1b[26~': Keys.F14,
    '\x1b[28~': Keys.F15,
    '\x1b[29~': Keys.F16,
    '\x1b[31~': Keys.F17,
    '\x1b[32~': Keys.F18,
    '\x1b[33~': Keys.F19,
    '\x1b[34~': Keys.F20,

    # Xterm
    '\x1b[1;2P': Keys.F13,
    '\x1b[1;2Q': Keys.F14,
    # '\x1b[1;2R': Keys.F15,  # Conflicts with CPR response.
    '\x1b[1;2S': Keys.F16,
    '\x1b[15;2~': Keys.F17,
    '\x1b[17;2~': Keys.F18,
    '\x1b[18;2~': Keys.F19,
    '\x1b[19;2~': Keys.F20,
    '\x1b[20;2~': Keys.F21,
    '\x1b[21;2~': Keys.F22,
    '\x1b[23;2~': Keys.F23,
    '\x1b[24;2~': Keys.F24,

    '\x1b[1;5A': Keys.ControlUp,     # Cursor Mode
    '\x1b[1;5B': Keys.ControlDown,   # Cursor Mode
    '\x1b[1;5C': Keys.ControlRight,  # Cursor Mode
    '\x1b[1;5D': Keys.ControlLeft,   # Cursor Mode

    '\x1b[1;2A': Keys.ShiftUp,
    '\x1b[1;2B': Keys.ShiftDown,
    '\x1b[1;2C': Keys.ShiftRight,
    '\x1b[1;2D': Keys.ShiftLeft,

    # Tmux sends following keystrokes when control+arrow is pressed, but for
    # Emacs ansi-term sends the same sequences for normal arrow keys. Consider
    # it a normal arrow press, because that's more important.
    '\x1bOA': Keys.Up,
    '\x1bOB': Keys.Down,
    '\x1bOC': Keys.Right,
    '\x1bOD': Keys.Left,

    '\x1b[5A': Keys.ControlUp,
    '\x1b[5B': Keys.ControlDown,
    '\x1b[5C': Keys.ControlRight,
    '\x1b[5D': Keys.ControlLeft,

    '\x1bOc': Keys.ControlRight,  # rxvt
    '\x1bOd': Keys.ControlLeft,  # rxvt

    # Tmux (Win32 subsystem) sends the following scroll events.
    '\x1b[62~': Keys.ScrollUp,
    '\x1b[63~': Keys.ScrollDown,

    '\x1b[200~': Keys.BracketedPaste,  # Start of bracketed paste.

    # Meta + arrow keys. Several terminals handle this differently.
    # The following sequences are for xterm and gnome-terminal.
    #     (Iterm sends ESC followed by the normal arrow_up/down/left/right
    #     sequences, and the OSX Terminal sends ESCb and ESCf for "alt
    #     arrow_left" and "alt arrow_right." We don't handle these
    #     explicitly, in here, because would could not distinguish between
    #     pressing ESC (to go to Vi navigation mode), followed by just the
    #     'b' or 'f' key. These combinations are handled in
    #     the input processor.)
    '\x1b[1;3D': (Keys.Escape, Keys.Left),
    '\x1b[1;3C': (Keys.Escape, Keys.Right),
    '\x1b[1;3A': (Keys.Escape, Keys.Up),
    '\x1b[1;3B': (Keys.Escape, Keys.Down),

    # Option+left/right on (some?) Macs when using iTerm defaults
    # (see issue #483)
    '\x1b[1;9D': (Keys.Escape, Keys.Left),
    '\x1b[1;9C': (Keys.Escape, Keys.Right),

    # Sequences generated by numpad 5. Not sure what it means. (It doesn't
    # appear in 'infocmp'. Just ignore.
    '\x1b[E': Keys.Ignore,  # Xterm.
    '\x1b[G': Keys.Ignore,  # Linux console.
}


def _get_reverse_ansi_sequences():
    """
    Create a dictionary that maps prompt_toolkit keys back to the VT100 escape
    sequences.
    """
    result = {}

    for sequence, key in ANSI_SEQUENCES.items():
        if not isinstance(key, tuple):
            if key not in result:
                result[key] = sequence

    return result


REVERSE_ANSI_SEQUENCES = _get_reverse_ansi_sequences()



"""
Search operations.

For the key bindings implementation with attached filters, check
`prompt_toolkit.key_binding.bindings.search`. (Use these for new key bindings
instead of calling these function directly.)
"""
from __future__ import unicode_literals
from .application.current import get_app
from .filters import is_searching, to_filter
from .key_binding.vi_state import InputMode
import six

__all__ = [
    'SearchDirection',
    'start_search',
    'stop_search',
]


class SearchDirection(object):
    FORWARD = 'FORWARD'
    BACKWARD = 'BACKWARD'

    _ALL = [FORWARD, BACKWARD]


class SearchState(object):
    """
    A search 'query', associated with a search field (like a SearchToolbar).

    Every searchable `BufferControl` points to a `search_buffer_control`
    (another `BufferControls`) which represents the search field. The
    `SearchState` attached to that search field is used for storing the current
    search query.

    It is possible to have one searchfield for multiple `BufferControls`. In
    that case, they'll share the same `SearchState`.
    If there are multiple `BufferControls` that display the same `Buffer`, then
    they can have a different `SearchState` each (if they have a different
    search control).
    """
    __slots__ = ('text', 'direction', 'ignore_case')

    def __init__(self, text='', direction=SearchDirection.FORWARD, ignore_case=False):
        assert isinstance(text, six.text_type)
        assert direction in (SearchDirection.FORWARD, SearchDirection.BACKWARD)

        ignore_case = to_filter(ignore_case)

        self.text = text
        self.direction = direction
        self.ignore_case = ignore_case

    def __repr__(self):
        return '%s(%r, direction=%r, ignore_case=%r)' % (
            self.__class__.__name__, self.text, self.direction, self.ignore_case)

    def __invert__(self):
        """
        Create a new SearchState where backwards becomes forwards and the other
        way around.
        """
        if self.direction == SearchDirection.BACKWARD:
            direction = SearchDirection.FORWARD
        else:
            direction = SearchDirection.BACKWARD

        return SearchState(text=self.text, direction=direction, ignore_case=self.ignore_case)


def start_search(buffer_control=None, direction=SearchDirection.FORWARD):
    """
    Start search through the given `buffer_control` using the
    `search_buffer_control`.

    :param buffer_control: Start search for this `BufferControl`. If not given,
        search through the current control.
    """
    from prompt_toolkit.layout.controls import BufferControl
    assert buffer_control is None or isinstance(buffer_control, BufferControl)
    assert direction in SearchDirection._ALL

    layout = get_app().layout

    # When no control is given, use the current control if that's a BufferControl.
    if buffer_control is None:
        if not isinstance(layout.current_control, BufferControl):
            return
        buffer_control = layout.current_control

    # Only if this control is searchable.
    search_buffer_control = buffer_control.search_buffer_control

    if search_buffer_control:
        buffer_control.search_state.direction = direction

        # Make sure to focus the search BufferControl
        layout.focus(search_buffer_control)

        # Remember search link.
        layout.search_links[search_buffer_control] = buffer_control

        # If we're in Vi mode, make sure to go into insert mode.
        get_app().vi_state.input_mode = InputMode.INSERT


def stop_search(buffer_control=None):
    """
    Stop search through the given `buffer_control`.
    """
    from prompt_toolkit.layout.controls import BufferControl
    assert buffer_control is None or isinstance(buffer_control, BufferControl)

    layout = get_app().layout

    if buffer_control is None:
        buffer_control = layout.search_target_buffer_control
        search_buffer_control = buffer_control.search_buffer_control
    else:
        assert buffer_control in layout.search_links.values()
        search_buffer_control = _get_reverse_search_links(layout)[buffer_control]

    # Focus the original buffer again.
    layout.focus(buffer_control)

    # Remove the search link.
    del layout.search_links[search_buffer_control]

    # Reset content of search control.
    search_buffer_control.buffer.reset()

    # If we're in Vi mode, go back to navigation mode.
    get_app().vi_state.input_mode = InputMode.NAVIGATION


def do_incremental_search(direction, count=1):
    """
    Apply search, but keep search buffer focused.
    """
    assert is_searching()
    assert direction in SearchDirection._ALL

    layout = get_app().layout

    search_control = layout.current_control
    prev_control = layout.search_target_buffer_control
    search_state = prev_control.search_state

    # Update search_state.
    direction_changed = search_state.direction != direction

    search_state.text = search_control.buffer.text
    search_state.direction = direction

    # Apply search to current buffer.
    if not direction_changed:
        prev_control.buffer.apply_search(
            search_state, include_current_position=False, count=count)


def accept_search():
    """
    Accept current search query. Focus original `BufferControl` again.
    """
    layout = get_app().layout

    search_control = layout.current_control
    target_buffer_control = layout.search_target_buffer_control
    search_state = target_buffer_control.search_state

    # Update search state.
    if search_control.buffer.text:
        search_state.text = search_control.buffer.text

    # Apply search.
    target_buffer_control.buffer.apply_search(search_state, include_current_position=True)

    # Add query to history of search line.
    search_control.buffer.append_to_history()

    # Stop search and focus previous control again.
    stop_search(target_buffer_control)


def _get_reverse_search_links(layout):
    """
    Return mapping from BufferControl to SearchBufferControl.
    """
    return dict((buffer_control, search_buffer_control)
            for search_buffer_control, buffer_control in layout.search_links.items())



"""
Implementation for async generators.

An asynchronous generator is one that can both produce `Future` objects as well
as actual values. These values have to be wrapped in a `AsyncGeneratorItem` in
order to be recognized. In the future, we can use the asynchronous generators
from Python 3 (and asyncio).

.. code:: python

    def async_generator():
        yield From(...)
        yield AsyncGeneratorItem(some_value)
        yield From(...)
        yield AsyncGeneratorItem(some_value)
        ...
"""
from __future__ import unicode_literals
from six.moves.queue import Queue
from threading import RLock
from .defaults import run_in_executor
from .future import Future
from .coroutine import From, Return

__all__ = [
    'AsyncGeneratorItem',
    'generator_to_async_generator',
    'consume_async_generator',
]


class AsyncGeneratorItem(object):
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return 'AsyncGeneratorItem(%r)' % (self.value, )


def generator_to_async_generator(get_iterable):
    """
    Turn a generator or iterable into an async generator.

    This works by running the generator in a background thread.
    The new async generator will yield both `Future` objects as well
    as the original items.

    :param get_iterable: Function that returns a generator or iterable when
        called.
    """
    q = Queue()
    f = Future()
    l = RLock()
    quitting = False

    def runner():
        """
        Consume the generator in background thread.
        When items are received, they'll be pushed to the queue and the
        Future is set.
        """
        for item in get_iterable():
            with l:
                q.put(item)
                if not f.done():
                    f.set_result(None)

            # When this async generator was cancelled (closed), stop this
            # thread.
            if quitting:
                break
        with l:
            if not f.done():
                f.set_result(None)

    # Start background thread.
    done_f = run_in_executor(runner, _daemon=True)

    try:
        while not done_f.done():
            # Wait for next item(s): yield Future.
            yield From(f)

            # Items received. Yield all items so far.
            with l:
                while not q.empty():
                    yield AsyncGeneratorItem(q.get())

                f = Future()

        # Yield final items.
        while not q.empty():
            yield q.get()

    finally:
        # When this async generator is closed (GeneratorExit exception, stop
        # the background thread as well. - we don't need that anymore.)
        quitting = True


def consume_async_generator(iterator, cancel, item_callback):
    """
    Consume an asynchronous generator.

    :param cancel: Cancel the consumption of the generator when this callable
        return True.
    :param item_callback: This will be called for each item that we receive.
    """
    assert callable(cancel)
    assert callable(item_callback)

    send = None
    try:
        item = iterator.send(send)
    except StopIteration:
        return

    while True:
        if cancel():
            break

        if isinstance(item, AsyncGeneratorItem):
            # Got item.
            item_callback(item.value)
            send = None

        elif isinstance(item, Future):
            # Process future.
            try:
                send = yield From(item)
            except BaseException as e:
                try:
                    item = iterator.throw(e)
                except StopIteration:
                    break
        else:
            raise TypeError('Expecting Completion or Future, got %r' % (item, ))

        try:
            item = iterator.send(send)
        except StopIteration:
            break

    raise Return(None)
